#!/usr/bin/perl
#
# (C) Hans-J. Barthen, DL5DI
#     Franz-Josef-Str. 20
#     D-56642 Kruft
#
# Copyright (c) 2008-2012 Hans-J. Barthen - DL5DI
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
#-----------------------------------------------------------------------------
# Revision history:
#
# 20111101: initial testversion
# 20111102: preparation for 2 step handling to database output, reduced decoding
# 20111103: included timeslot
# 20111104: fix gid issue (EU/NA)
# 20111107: VIS-switch and flags implemented
# 20111108: user_name and rptr_city added, switch port to UDP port
# 20111109: peer authentication by IP address
# 20111113: get more raw data, move interpretation from client to server
# 20111115: flag decoding restructured
# 20111130: create network list
# 20111202: bugfix network list
# 20111205: use rptr-id for reporting peers, change user-db to sql
# 20111230: switch to receiver for tcpdump/windump -Tdmr
# 20111231: portnumber implemented to run different instances
# 20120101: portnumber moved to command line argument
# 20120102: multiport-version, portnumber in datastream
# 20120114: bugfix individual call (get destination from user-table)
# 20120211: take rptrcall from sql-db if not found in preloaded list, debug switch added to commandline
# 20120215: re-open database connection if it was closed 
# 20120303: don't overwrite rptrcall in rptr-act table with empty input from list
# 20120320: added OH/Finland on port 50006
# 20120329: added DL-West on port 50007
# 20120331: added Master-Master-keep-alive
# 20120401: bugfix Master-Master-keep-alive
# 20120413: use rptr data from SQL, no longer table in ram
# 20120421: new talkgroup structure
# 20120608: change of input format
# 20120630: cleanup

$version = "20120630";

use IO::Socket::INET;
use DBI;

# flush after every write
$| = 1;

# debug = 0 -> off
# debug = 1 -> load arrays and exit
# debug = 2 -> show sql commands
# debug = 3 -> output
# debug = 4 -> link status messages
# debug = 5 -> input
# debug = 6 -> raw data
# debug = 0;
$debug = 0;

if (!$ARGV[0]) {
    print "Error! Missing Portnumber to listen to!\n";
    exit;
} elsif ($ARGV[0] eq "-d"){
    $debug = 3;
    if(!$ARGV[1]) {
	print "Error! Missing Portnumber to listen to!\n";
	exit;
    } else {
	$port = $ARGV[1];
    }
} else {
    $port = $ARGV[0];
}

$port =~ y/a-zA-Z//;
if($port == 0) {
    print "Error! Portnumber out of range (1-65535)!\n";
    exit;
}

#-----------------------------------------------------------------------------
# peer_id stored to the database to mark origin of data
#-----------------------------------------------------------------------------
# DL-Master / SmartPTT
$peer_id = "262401";
#
#-----------------------------------------------------------------------------
# network (use portnumber where it is connected)
# status of 20120414:
#
# 50000: MARC4 DL-Master
# 50050: Wien
# 50002: MARC3 Schaumburg
# 50003: (Test Walter)
# 50004: MARC2 Florida
# 50005: MARC1 New York
# 50006: Finnland
# 50007: DL-West Frankfurt
# 50008:   ?USA?
#
#-----------------------------------------------------------------------------
# this destination ID will switch visivility of the source ID on and off
# (data privacy)

$vis_on = 26200;
$vis_off = 26211;

#-----------------------------------------------------------------------------

my ($socket,$rdata);
my ($peeraddress,$peerport);
my $database = "DMR";
my $user = "dmruser";
my $pw = "egal";
my $host = "127.0.0.1";
my $table1 = "LastHeard";
my $table2 = "Log";
my $table3 = "ActRptr";
my $table4 = "dmr_usr";

$dmrdata = "/opt/dmr/data/";
$rptrfile = $dmrdata . "dmr_rptr.txt";
$gidfile1 = $dmrdata . "dmr_gid1.txt";
$gidfile2 = $dmrdata . "dmr_gid2.txt";
$logfile = "/var/log/dmr/dmr-monitor.log";

# index = portnumber - 50000
$netname[0] = "4"; # "DE WEST";
$netname[1] = "50"; # "AT";
$netname[2] = "3"; # "US-IL";
$netname[3] = "8"; # "CH";
$netname[4] = "2"; # "US-FL"
$netname[5] = "1"; # "US-NY"
$netname[6] = "6"; # "FI"
$netname[7] = "7"; # "DE MITTE"
$netname[8] = "9"; # "?USA?"
$netname[9] = "10"; # "?USA?"

#-----------------------------------------------------------------------------
# read talkgroup destinations fo TS1 into table
#-----------------------------------------------------------------------------

open(DESTDATA, $gidfile1) or die "cannot open < input.txt: $!";
while(<DESTDATA>){
    $destin = $_;
    chop($destin);
    $destin =~ s/ \"//g;
    $destin =~ s/\"//g;
    if(index($destin,'#') == 0) { next;}
    if(length($destin) > 6){
	($destid1,$net1,$destname1,$junk1) = split(/,/,$destin,4);
	if($debug == 1){
	    print "$destid1 - $net1 - $destname1\n";
	}
	$Grouplist1[$destid1][$net1] = $destname1;
    }
}
close(DESTDATA);

#-----------------------------------------------------------------------------
# read talkgroup destinations fo TS2 into table
#-----------------------------------------------------------------------------

open(DESTDATA, $gidfile2) or die "cannot open < input.txt: $!";
while(<DESTDATA>){
    $destin = $_;
    chop($destin);
    $destin =~ s/ \"//g;
    $destin =~ s/\"//g;
    if(index($destin,'#') == 0) { next;}
    if(length($destin) > 6){
	($destid2,$net2,$destname2,$junk) = split(/,/,$destin,4);
	if($debug == 1){
	    print "$destid2 - $net2 - $destname2\n";
	}
	$Grouplist2[$destid2][$net2] = $destname2;
    }
}
close(DESTDATA);

#-----------------------------------------------------------------------------
# save start time to logfile
#-----------------------------------------------------------------------------
($sec,$min,$hour,$day,$mon,$year,$wday,$yday,$isdst)=gmtime(time);
$mon++;
$year+=1900;
$timestamp = sprintf("%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d",$year,$mon,$day,$hour,$min,$sec);

open(LOGFILE, ">>$logfile") or die "cannot open < logfile: $!";
printf LOGFILE "$timestamp Program Start - Version $version listening on port $port\n";
close(LOGFILE);

#-----------------------------------------------------------------------------

if($debug == 1){
    exit;
}

$socket = new IO::Socket::INET (LocalPort => $port,Proto => 'udp',) or die "ERROR in Socket Creation : $!\n";
#$dbh = DBI->connect("DBI:mysql:database=$database:host=$host", $user) or die "Can't connect to database: $DBI::errstr\n";

$err = 0;

#
# Old format of input string:
# 2011-11-13 10:52:04 50000 84 00 262440 139 2624420 2624415 01 00 00 2e 66 20 80 dd 01 65 74 54 d5 67 53 00 00 00
# date       time     port  o0 o1 rptr_id s  src_id  dest_id o12.................o19 020.......................029
#
# New format of input string:
# 2012-06-05 21:45:09 50000 80 262510 237 2626010 2624415 02 00006da0 20 80 dd 50820 0f5cc863 0 01 80 11 84 0a 96
# date       time     port  o0 rptr_id s  src_id  dest_id o12 flctrl  o17o18o19 seq  timestmp s o30o31 l 34 35 dz
#
# s: sequence number
# o#: octed number (octed 0-12 DMR = header, octed 20-29 = appended data : TCP/IP header, GPS etc.)
#

while($err == 0) {

    $dbh ||= DBI->connect("DBI:mysql:database=$database:host=$host", $user) or die "Can't connect to database: $DBI::errstr\n";

#    $socket->recv($frame,1024);
#    $peer_address = $socket->peerhost();
#    $peer_port = $socket->peerport();

#    $client = getpeername($socket) or die "Couldn't identify other end: $!\n";;
#    ($peer_port,$peer_paddr) = unpack_sockaddr_in($client);
#    $peer_address = inet_ntoa($peer_paddr);

    $frame = $socket->getline();

    if($debug == 6){
	printf "(%s:%s)\n",$peer_address,$peer_port;
	print $frame;
    }

    ($lhdate,$lhtime,$net,$rawdata) = split(/ /,$frame,4);
    chomp($frame);
    $lastheard = $lhdate . " " . $lhtime;

    if(index($frame,"windump") > 0){
	($sec,$min,$hour,$day,$mon,$year,$wday,$yday,$isdst)=gmtime(time);
	$mon++;
	$year+=1900;

	open(LOGFILE, ">>$logfile") or die "cannot open < logfile: $!";
	printf(LOGFILE "\n%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d Client started on port $port:\n  %s\n",$year,$mon,$day,$hour,$min,$sec,$frame);
	close(LOGFILE);
	if($debug > 0){
	    printf("\n%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d Client started:\n  %s\n",$year,$mon,$day,$hour,$min,$sec,$frame);
	}
    }

    if($debug == 5){
	printf "raw: (%s/%s)\n<%s>\n",$peer,$peer_id,$frame;
    }

    if((substr($rawdata,0,1)) eq "8"){

# 80 262510 237 2626010 2624415 02 00006da0 20 80 dd 50820 0f5cc863 0 01 80 11 84 0a 96
# o0 rptr_id s  src_id  dest_id o12 flctrl  o17o18o19 seq  timestmp s o30o31 l 34 35 dz

	($oct0,$rptr_id,$seq,$src_id,$dest_id,$oct12,$flctl,$oct17,$oct18,$oct19,$seqnum,$tstamph,$sync,$o30,$o31,$words,$o34,$o35,$dsize) = split(/ /,$rawdata,19);
	if($debug == 3){
	    printf "(%s) %s %s\n",$peer,$peer_id,$frame;
	}

	if($dest_id == 2442000){ next; }
	if($src_id == 2442000){ next; }

	if($src_id ne $src_id_last){
	    $sql = "SELECT `call`, `name` FROM `$table4` WHERE `user_id` = '$src_id'";
	    if($debug == 2){
		print $sql . "\n";
	    }
	    $sth = $dbh->prepare($sql);
	    $num = $sth->execute();
	    if($sth->rows() > 0){
		@data = $sth->fetchrow_array;
		$src_call = $data[0];
		$fromname = $data[1];
	    } else {
		$src_call = "";
		$fromname = "";
	    }
	    $src_id_last = $src_id;
	}

#----- Switch VIS-Flag:
	$vis = -1;
	if($dest_id == $vis_off){
	    $vis = 0; 			# VIS OFF
	} elsif($dest_id == $vis_on){ 
	    $vis = 1; 			# VIS ON
	}

#----- decode timeslot and F-flag  from octet 17:
        $ts = "1";
        $oct17h = substr($oct17,0,1);
        if(hex($oct17h) & 2) { $ts = "2"; }
        if(hex($oct17h) & 4) {
	    $flags .= "F";
	    $pttend = $oct17h & 4;
	}
	$t = $net-50000;
	$n = $netname[$t];
	$group = "";
	if($ts == "1"){
	    if($Grouplist1[$dest_id][$n]){
		$group = $Grouplist1[$dest_id][$n];
	    } elsif($Grouplist1[$dest_id][0]){
		$group = $Grouplist1[$dest_id][0];
	    }
	} else {
	    if($Grouplist2[$dest_id][$n]){
		$group = $Grouplist2[$dest_id][$n];
	    } elsif($Grouplist2[$dest_id][0]){
		$group = $Grouplist2[$dest_id][0];
	    }
	}
	if(length($group) == 0){
	    $sql = "SELECT `call` FROM `$table4` WHERE `user_id` = '$dest_id'";
	    if($debug == 2){
		print $sql . "\n";
	    }
	    $sth = $dbh->prepare($sql);
	    $num = $sth->execute();
	    if($sth->rows() > 0){
		@data = $sth->fetchrow_array;
		$group = $data[0];
	    } else {
		$group = "";
	    }
	}
	$sql = "select rptr_call,city FROM $table3 where rptr_id='$rptr_id'";
	$sth = $dbh->prepare($sql);
	$num = $sth->execute();
	if($sth->rows() == 0){
		$rptrcall = "";
		$rptrcity = "";
	} else {
	    while (my @data = $sth->fetchrow_array) {
		$rptrcall = $data[0];
		$rptrcity = $data[1];
	    }
	}

       $flags = "";

#----- decode A-, G-, I-flags from octet 1:
#	$agi = substr($oct0,1,1);
#	$flags = $flag{$agi};
        $oct0l = substr($oct0,1,1);
	if(hex($oct0l) & 4) {
	    $flags .= "A";
	} elsif(hex($oct0l) & 1) { 
	    $flags .= "I";
	} else { 
	    $flags .= "G";
	}

#----- decode D-, V-flag from octet 12:
        $oct12l = substr($oct12,1,1);
	if(hex($oct12l) & 1) { $flags .= "D";}
	if(hex($oct12l) & 2) { $flags .= "V";}

#----- decode rawdata-flag from octet 19:
        $oct19h = substr($oct19,0,1);
	if(hex($oct19h) & 8) { $flags .= "R";}

#	Should dispatcher transmission be visible?
	$hidden = 0;
	if($src_id < 100000) { $hidden = 1; }

#	Sysop destination
	if($dest_id == "262007") { $hidden = 1; }

#	RSSI Flag
        $o31h = substr($o31,0,1);
	if(hex($o31h) & 4) { $rssi = 0; } else { $rssi = 1; }

#----- MHeard-Table:
	$sql = "select * FROM $table1 where from_id='$src_id'";
	$sth = $dbh->prepare($sql);
	$num = $sth->execute();
	if($sth->rows() > 0){
	    if($vis < 0){
		$sql = "UPDATE $table1 SET lastdate='$lastheard',from_call='$src_call',from_name='$fromname',rptr_id='$rptr_id',rptr_call='$rptrcall',rptr_city='$rptrcity',grp_id='$dest_id',grp_name='$group',ts='$ts',flags='$flags',hidden='$hidden',net='$net',rssi='$rssi' WHERE from_id='$src_id';";
	    } else {
		$sql = "UPDATE $table1 SET lastdate='$lastheard',from_call='$src_call',from_name='$fromname',rptr_id='$rptr_id',rptr_call='$rptrcall',rptr_city='$rptrcity',grp_id='$dest_id',grp_name='$group',ts='$ts',flags='$flags',vis='$vis',hidden='$hidden',net='$net',rssi='$rssi' WHERE from_id='$src_id';";
	    }
	} else {
	    if($vis < 0){
        	$sql = "INSERT INTO $table1 (lastdate,from_id,from_call,from_name,rptr_id,rptr_call,rptr_city,grp_id,grp_name,ts,flags,hidden,net,rssi) VALUES ('$lastheard','$src_id','$src_call','$fromname','$rptr_id','$rptrcall','$rptrcity','$dest_id','$group','$ts','$flags','$hidden','$net','$rssi');";
	    } else {
        	$sql = "INSERT INTO $table1 (lastdate,from_id,from_call,from_name,rptr_id,rptr_call,rptr_city,grp_id,grp_name,ts,flags,vis,hidden,net,rssi) VALUES ('$lastheard','$src_id','$src_call','$fromname','$rptr_id','$rptrcall','$rptrcity','$dest_id','$group','$ts','$flags','$vis','$hidden','$net','$rssi');";
	    }
	}
	if($debug == 2){
	    print $sql ."\n";
	}
	$sth = $dbh->prepare($sql);
	$sth->execute();

#----- Log-Table:
	$sql = "select * FROM $table2 where date_time> UTC_TIMESTAMP() - INTERVAL 15 MINUTE AND from_id='$src_id' AND rptr_id='$rptr_id' AND seq='$seq'";
	$sth = $dbh->prepare($sql);
	$num = $sth->execute();
	if($sth->rows() == 0){
	    $sql = "INSERT INTO $table2 (date_time,from_id,from_call,from_name,rptr_id,rptr_call,rptr_city,grp_id,grp_name,seq,ts,flags,hidden,peer,end_time,rawdata,source,net,rssi,words,seqnum) VALUES ('$lastheard','$src_id','$src_call','$fromname','$rptr_id','$rptrcall','$rptrcity','$dest_id','$group','$seq','$ts','$flags','$hidden','$peer','$lhtime','$rawdata','$peer_id','$net','$rssi','$words','$seqnum');";
	    if($debug == 2){
		print $sql ."\n";
	    }
	    $sth = $dbh->prepare($sql);
	    $sth->execute();
	} elsif($pttend){
	    $sql = "UPDATE $table2 SET end_time='$lhtime',flags='$flags' where from_id='$src_id' AND rptr_id='$rptr_id' AND seq='$seq' AND net='$net'";
	    if($debug == 2){
		print $sql ."\n";
	    }
	    $sth = $dbh->prepare($sql);
	    $sth->execute();
	}

    } elsif(((substr($rawdata,0,2)) eq "96") || ((substr($rawdata,0,2)) eq "97") || ((substr($rawdata,0,2)) eq "98") || ((substr($rawdata,0,2)) eq "99")){
#
# Format of input string:
# 2011-11-13 10:52:04 98 262440 6a
# date       time     o0 rptrid o5
#
# c:  client-id
# o5: octed5 showing link status of timeslots
#
# 96: Master-keep-alive-request
# 97: Master-keep-alive-response
# 98: Peer-keep-alive-request
# 99: Peer-keep-alive-response
#
	($oct0,$rptr_id,$ts,$rest) = split(/ /,$rawdata,4);
	if(($debug == 3) || ($debug == 4)){
	    printf "(%s) %s %s\n",$peer,$peer_id,$frame;
	}

	$sql = "select rptr_call,city FROM $table3 where rptr_id='$rptr_id'";
	$sth = $dbh->prepare($sql);
	$num = $sth->execute();
	if($sth->rows() == 0){
		$rptrcall = "";
		$rptrcity = "";
	} else {
	    while (my @data = $sth->fetchrow_array) {
		$rptrcall = $data[0];
		$rptrcity = $data[1];
	    }
	}
	$master=0;
	if(($oct0 eq "96") || ($oct0 eq "97")){
	    $master=1;
	}
	$ts1 = "0";
	$ts2 = "0";
        $tsh = substr($ts,1,1);
        if(hex($tsh) & 2) { $ts2 = "1"; }
        if(hex($tsh) & 8) { $ts1 = "1"; }

#----- RptrAct-Table:
	$sql = "select * FROM $table3 where rptr_id='$rptr_id'";
	$sth = $dbh->prepare($sql);
	$num = $sth->execute();
	if($sth->rows() == 0){
	    $sql = "INSERT INTO $table3 (lastheard,rptr_id,rptr_call,ts1,ts2,net,master) VALUES ('$lastheard','$rptr_id','$rptrcall','$ts1','$ts2','$net','$master');";
	} else {
	    if(length($rptrcall) > 0){
		$sql = "UPDATE $table3 SET lastheard='$lastheard',rptr_call='$rptrcall',ts1='$ts1',ts2='$ts2',net='$net',master='$master' where rptr_id='$rptr_id';";
	    } else {
		$sql = "UPDATE $table3 SET lastheard='$lastheard',ts1='$ts1',ts2='$ts2',net='$net',master='$master' where rptr_id='$rptr_id';";
	    }
	}
	if($debug == 2){
	    print $sql ."\n";
	}
	$sth = $dbh->prepare($sql);
	$sth->execute();
    }
}
$dbh->close;
$socket->close();

exit;
